﻿''' <summary>
''' bInterface class
''' 
''' 
''' The bInterface class is a Visual Basic .NET class that can be used to control the
''' RS-232 computer interface I've designed.
''' 
''' This class is still in an experimental phase: some bugs might still be present.
''' You shouldn't use it for any critical applications yet.
''' 
''' </summary>
''' <remarks>Copyright (c) 2012 - Vanderhaegen Bart</remarks>
Public Class BVinterface

    ' General class properties
    Private Const MAX_SHIFT_REGISTERS = 15
    Private Const MAX_DATA_PACKETS As Integer = 15
    Private Const BAUD As Integer = 9600

    ' Private class members for internal use
    Private _portNumber As String
    Private _shiftRegisters As Integer
    Private _isConnected As Boolean
    Private _initialized As Integer
    Private _sendImmediately As Boolean = False
    Private _shiftBitOnefirst As Boolean = True
    Private _shiftByteOneFirst As Boolean = True

    ' Internal data structure
    ' Its size will depend on the desired number of shift registers
    Private databytes(0) As Byte
    ' define our class wide COM port control

    Private WithEvents comCtrl As New System.IO.Ports.SerialPort

    ' Public events that can be used by the users of this class
    Public Event feedbackReceived(ByVal message As String)





#Region "Class Constructor"
    ''' <summary>
    ''' The default class constructor without any parameters
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub New()
    End Sub
#End Region





#Region "Class Properties"
    ''' <summary>
    ''' The shiftBitOneFirst() affects how the bits are passed to the interface card's shift registers.
    ''' 
    ''' shiftBitOneFirst = true: 
    ''' The channel 1 of a chip will correspond with the output QH.
    ''' 
    ''' shiftBitOneFirst = false:
    ''' The channel 1 of a chip will correspond with the output QA.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property shiftBitOneFirst() As Boolean
        Get
            Return _shiftBitOnefirst
        End Get
        Set(value As Boolean)
            If isConnected Then
                Throw New Exception("You cannot change this option when the interface is connected")
            Else
                _shiftBitOnefirst = value
            End If
        End Set
    End Property



    ''' <summary>
    ''' The shiftByteOneFirst() affects how the bytes are passed to the interface card's shift registers.
    ''' 
    ''' shiftByteOneFirst = true: 
    ''' The channel 1 will appear at the outputs of the last cascaded shift register.
    ''' 
    ''' shiftByteOneFirst = false:
    ''' The channel 1 will appear at the outputs of the shift register that is connected to the
    ''' PIC microcontroller (the very first register).
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property shiftByteOneFirst As Boolean
        Get
            Return _shiftByteOneFirst
        End Get
        Set(value As Boolean)
            If isConnected Then
                Throw New Exception("You cannot change this option when the interface is connected")
            Else
                _shiftByteOneFirst = value
            End If
        End Set
    End Property



    ''' <summary>
    ''' comPort is used to assign the desired COM portnumber to the class.
    ''' </summary>
    ''' <value>The number of the COM port, in a string format. This string should include the word COM. Ex: "COM1"</value>
    ''' <returns>The COM port number that was already assigned</returns>
    ''' <remarks>You cannot change the COM port number when the interface is already connected.</remarks>
    Public Property comPort As String
        Get
            Return _portNumber
        End Get
        Set(value As String)
            If Not isConnected Then
                _portNumber = value
            Else
                Throw New Exception("You cannot change the COM port while the interface is connected")
            End If
        End Set
    End Property



    ''' <summary>
    ''' sendImmediately indicates the data is sent immediately after one of the bits is changed. During
    ''' each SEND operation ALL of the bytes are transmitted to the interface board.
    ''' </summary>
    ''' <value>A boolean to indicate if you want to activate the sendImmediately or not.</value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property sendImmediately As Boolean
        Get
            Return _sendImmediately
        End Get
        Set(value As Boolean)
            _sendImmediately = value
        End Set
    End Property



    ''' <summary>
    ''' isConnected() returns true if the interface is connected. Otherwise it returns false. No real
    ''' communication check is performed, the function just checks if the COM port was opened or not.
    ''' </summary>
    ''' <value></value>
    ''' <returns>Returns true when the COM port is opened</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isConnected() As Boolean
        Get
            Return _isConnected
        End Get
    End Property


    ''' <summary>
    ''' isInitialized() returns true when the class is initialized. This is the case when the number
    ''' of shift registers is specified.
    ''' </summary>
    ''' <value></value>
    ''' <returns>A boolean to indicate if the class is initialized</returns>
    ''' <remarks></remarks>
    Public ReadOnly Property isInitialized As Boolean
        Get
            Return _initialized
        End Get
    End Property

#End Region





#Region "RS-232 Communications"
    ''' <summary>
    ''' Connect() will open the selected COM port. The function will set all connection-specific
    ''' parameters.
    ''' </summary>
    ''' <returns>Returns true when the delected COM port could be opened</returns>
    ''' <remarks>You need to call the COMport property first</remarks>
    Public Function connect() As Boolean
        If isInitialized Then
            Try
                With comCtrl
                    .PortName = comPort
                    .Parity = IO.Ports.Parity.None
                    .BaudRate = BAUD
                    .DataBits = 8
                    .DtrEnable = False
                    .StopBits = IO.Ports.StopBits.One
                End With

                comCtrl.Open()
                _isConnected = True
            Catch ex As Exception
                _isConnected = False
            End Try

        Else
            Throw New Exception("The interface is not initialized yet!")
        End If

        Return isConnected()
    End Function



    ''' <summary>
    ''' disconnect() will close the COM port.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub disconnect()
        comCtrl.Close()
        _isConnected = False
    End Sub



    ''' <summary>
    ''' comPort_DataReceived is an event that is called when some incoming RS-232 data is received. The event
    ''' will call the feedbackReceived() event, an event that can be used by the end-user. This feedbackReceived
    ''' event will be called with the parameter "OK" or "ERR".
    ''' 
    ''' An OK will be transmitted when the request packet was understood by the interface.
    ''' An ERR packet will be transmitted in the case of an error: communication or checksum error.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub comPort_DataReceived(sender As Object, e As System.IO.Ports.SerialDataReceivedEventArgs) Handles comCtrl.DataReceived

        While comCtrl.BytesToRead > 0
            Dim receivedbyte As Byte = comCtrl.ReadByte()

            Select Case receivedbyte
                Case &HF
                    RaiseEvent feedbackReceived("OK")
                Case &HF0
                    RaiseEvent feedbackReceived("ERR")
                Case Else
                    RaiseEvent feedbackReceived("Unknown response:" & receivedbyte)
            End Select
        End While
    End Sub



    ''' <summary>
    ''' transmitData() will transmit the actual data. The interface must be connected before this
    ''' method can be called. Otherwise an exception will be raised.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub transmitData()

        If isConnected() Then
            databytes(0) = &HF0             ' fill in start byte
            databytes(1) = shiftRegisters   ' fill in number of shift registers
            addCheckSum()                   ' adds the checksum

            Try
                Debug.Print("writing " & UBound(databytes) + 1 & " bytes to port!")
                comCtrl.Write(databytes, 0, UBound(databytes) + 1)
            Catch ex As Exception

            End Try

        Else
            Throw New Exception("COM port isn't opened yet!")
        End If
    End Sub


#End Region





#Region "Digital Output Controlling & Handlers"
    ''' <summary>
    ''' The shiftRegisters property can be used to set the number of shift registers. Each shift register will 
    ''' represent 8 output bits. You can select 1 to 15 shift registers (8 to 120 digital outputs).
    ''' 
    ''' Total outputs = shiftRegisters * 8
    ''' </summary>
    ''' <value>The number of shift registers you want to use.</value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property shiftRegisters As Integer
        Get
            Return _shiftRegisters
        End Get
        Set(value As Integer)
            If (value > 0 And value <= MAX_SHIFT_REGISTERS) Then
                ReDim databytes(value + 2)  ' redimension the array
                _shiftRegisters = value     ' set number of registers
                _initialized = True         ' set initialized to true
            Else
                Throw New Exception("please select 1 to " & MAX_SHIFT_REGISTERS & " shift registers!")
            End If
        End Set
    End Property



    ''' <summary>
    ''' The output property will set or clear a certain bit on the interface. You can set or clear any
    ''' existing channel. The channel number must be between 1 and (8 * shiftRegisters), otherwise an
    ''' error is generated.
    ''' </summary>
    ''' <param name="channel">The channel to set or clear</param>
    ''' <value>The desired output state. True=ON, False=OFF</value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Property output(channel As Integer) As Boolean
        Get
            ' Get the desired bit
            Dim chip, bit As Integer
            channelPosition(channel, chip, bit)
            Return (databytes(chip + 2) And Math.Pow(2, bit)) > 0

        End Get
        Set(value As Boolean)
            ' Set the desired bit
            Dim chip, bit As Integer
            channelPosition(channel, chip, bit)

            If value Then
                databytes(chip + 2) = databytes(chip + 2) Or Math.Pow(2, bit)
            Else
                databytes(chip + 2) = databytes(chip + 2) And Not Math.Pow(2, bit)
            End If

            ' Send the data to the interface
            If sendImmediately Then
                transmitData()
                Debug.Print("Sending: " & ByteArrayToString(databytes))
            End If
        End Set
    End Property



    ''' <summary>
    ''' The setOutput() method will activate one of the digital outputs.
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <remarks></remarks>
    Public Sub setOutput(ByVal channel As Integer)
        output(channel) = True
    End Sub



    ''' <summary>
    ''' The clearOutput() method will clear one of the digital outputs.
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <remarks></remarks>
    Public Sub clearOutput(ByVal channel As Integer)
        output(channel) = False

    End Sub



    ''' <summary>
    ''' the InvertOutput() method will invert one of the digital outputs.
    ''' 
    ''' The output state is toggled during an invertion.
    ''' </summary>
    ''' <param name="channel"></param>
    ''' <remarks></remarks>
    Public Sub invertOutput(ByVal channel As Integer)
        output(channel) = Not output(channel)
    End Sub



    ''' <summary>
    ''' The clearAllOutputs() method will clear all digital outputs.
    ''' 
    ''' The channels are only updated when the sendImmediately property is set to true. Otherwise
    ''' you need to call the transmitData() method yourself.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub clearAllOutputs()
        Dim prevParam As Boolean = sendImmediately
        sendImmediately = False

        For channel As Integer = 1 To (8 * shiftRegisters)
            output(channel) = False
        Next

        sendImmediately = prevParam
        If sendImmediately Then transmitData()
    End Sub



    ''' <summary>
    ''' The setAllOutputs() method will set all digital outputs.
    ''' 
    ''' The channels are only updated when the sendImmediately property is set to true. Otherwise
    ''' you need to call the transmitData() method yourself.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub setAllOutputs()
        Dim prevParam As Boolean = sendImmediately
        sendImmediately = False

        For channel As Integer = 1 To (8 * shiftRegisters)
            output(channel) = True
        Next

        sendImmediately = prevParam
        If sendImmediately Then transmitData()
    End Sub



    ''' <summary>
    ''' The invertAllOutputs() method will set all digital outputs.
    ''' 
    ''' The channels are only updated when the sendImmediately property is set to true. Otherwise
    ''' you need to call the transmitData() method yourself.
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub invertAllOutputs()
        Dim prevParam As Boolean = sendImmediately
        sendImmediately = False

        For channel As Integer = 1 To (8 * shiftRegisters)
            output(channel) = Not output(channel)
        Next

        sendImmediately = prevParam
        If sendImmediately Then transmitData()
    End Sub



    ''' <summary>
    ''' The channelPosition() method is an aider function that will calculate the position of
    ''' the desired chip and bit.
    ''' 
    ''' The properties shiftByteOneFirst and shiftBitOneFirst will affect the bit/byte order.
    ''' 
    ''' The chip and bit variables are defined as byRef, so they are passed back to the caller
    ''' function.
    ''' </summary>
    ''' <param name="channel">The channel to find</param>
    ''' <param name="chip">The matching chip</param>
    ''' <param name="bit">The matching bit</param>
    ''' <remarks></remarks>
    Private Sub channelPosition(ByVal channel As Integer, ByRef chip As Integer, ByRef bit As Integer)

        If isInitialized Then
            If channel > 0 And channel < (1 + 8 * shiftRegisters) Then
                channel = channel - 1
                chip = channel \ 8   ' the number of the shift register (zero based index)
                bit = channel Mod 8  ' the number of the output bit     (zero based index)

                If shiftByteOneFirst = False Then
                    chip = shiftRegisters - chip - 1
                End If


                If shiftBitOneFirst = False Then
                    bit = 7 - bit
                End If


            Else
                Throw New Exception("please select a channel between 1 and " & (8 * shiftRegisters))
            End If
        Else
            Throw New Exception("The interface is not initialized yet. Call the setNumberOfRegisters property first!")
        End If

    End Sub



    ''' <summary>
    ''' addCheckSum() will add a checksum to the existing data
    ''' </summary>
    ''' <remarks></remarks>
    Public Sub addCheckSum()
        Dim sum As Integer = 0

        For byteCnt As Integer = 0 To UBound(databytes) - 1
            sum = sum + databytes(byteCnt)
        Next
        databytes(UBound(databytes)) = sum Mod 256
    End Sub
#End Region





#Region "Aider Functions"
    ''' <summary>
    ''' The ByteArrayToString() will show the contents as the byte array as a string.
    ''' </summary>
    ''' <param name="byteArray"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Function ByteArrayToString(ByVal byteArray() As Byte) As String
        Return BitConverter.ToString(byteArray)
    End Function
#End Region

End Class
